/*
 Copyright (c) 2012 Michael Zucchi

 This file is part of soclesdemo, test and demo routines
 for socles, an OpenCL image processing library.

 soclesdemo is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 soclesdemo is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with soclesdemo.  If not, see <http://www.gnu.org/licenses/>.
 */
package soclesdemo.feature;

import au.notzed.socle.SOCLEContext;
import au.notzed.socle.feature.surf.SURFDetector;
import au.notzed.socle.filter.ConvolveXY;
import au.notzed.socle.image.ImageFormats;
import au.notzed.socle.ransac.RANSACHomography;
import au.notzed.socle.util.ImageLoader;
import com.jogamp.opencl.*;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import static java.lang.Math.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.SwingUtilities;
import soclesdemo.SoclesDemo;

/**
 * SURF + RANSACFeatures demo
 *
 * This finds features in two images, finds putative matches using nearest-neighbour matching,
 * and then uses RANSACFeatures to discard outliers and generate a homography (transform matrix) mapping
 * one set of points to the other.
 *
 * @author notzed
 */
public class RANSACFeatures extends SoclesDemo {

	public static void main(final String[] args) throws InterruptedException {
            Thread t;
            for(int i=0; i<20; i++){
                t = new Thread(){
                        @Override
			public void run() {
				try {
					String[] a = args;
					if (a.length == 0)
						a = new String[]{"D:\\Documents\\images\\Testing\\Images\\form3.jpg",
							"D:\\Documents\\images\\Testing\\Models\\form1.jpg"};

					if (a.length < 2) {
						System.err.println("Specify two images to compare");
						System.exit(1);
					}

					SOCLEContext sc = new SOCLEContext(CLContext.create(CLDevice.Type.GPU));
					CLContext cl = sc.cl;
					CLCommandQueue q = sc.getQ();

					BufferedImage img0 = ImageIO.read(new File(a[0]));
					BufferedImage img1 = ImageIO.read(new File(a[1]));
					int id = 0;
					int width = img0.getWidth();
					int height = img0.getHeight();

					img1 = sameSize(img1, img0);

					SURFDetector surf = new SURFDetector(cl, 2, 1024 * 3, width, height, 2, 0.0001f);

					CLImage2d<ByteBuffer> clsrc0 = cl.createImage2d(ByteBuffer.allocateDirect(width * height).order(ByteOrder.nativeOrder()), width, height, ImageFormats.R_UINT8);
					CLImage2d<ByteBuffer> clsrc1 = cl.createImage2d(ByteBuffer.allocateDirect(width * height).order(ByteOrder.nativeOrder()), width, height, ImageFormats.R_UINT8);

					ImageLoader.putImage(q, img0, clsrc0);
					ImageLoader.putImage(q, img1, clsrc1);

					// Get features for both images
					surf.loadImage(q, 0, clsrc0);
					surf.loadImage(q, 1, clsrc1);

					System.out.println("Point count 0 = " + surf.getPointCount(0));
					System.out.println("Point count 1 = " + surf.getPointCount(1));

					// Find putative matches
					int pointCount = surf.getPointCount(0);

					CLBuffer<FloatBuffer> bestDistance = cl.createFloatBuffer(pointCount);
					CLBuffer<IntBuffer> bestIndex = cl.createIntBuffer(pointCount);
					CLBuffer<FloatBuffer> m2permute = cl.createFloatBuffer(pointCount * 2);

					surf.nearestMatch(q, 0, 1, bestDistance, bestIndex, m2permute);

					// putative matches
					if (false) {
						q.putReadBuffer(m2permute, false);
						q.putReadBuffer(surf.getPoints(0), true);

						BufferedImage tmp = new BufferedImage(width * 2, height, BufferedImage.TYPE_INT_BGR);
						Graphics2D gg = tmp.createGraphics();

						gg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
						gg.drawImage(img0, 0, 0, null);
						gg.drawImage(img1, width, 0, null);

						Line2D.Float line = new java.awt.geom.Line2D.Float();

						for (int i = 0; i < pointCount; i++) {
							line.x1 = surf.getPoints(0).getBuffer().get(i * 2 + 0);
							line.y1 = surf.getPoints(0).getBuffer().get(i * 2 + 1);
							line.x2 = m2permute.getBuffer().get(i * 2 + 0) + width;
							line.y2 = m2permute.getBuffer().get(i * 2 + 1);

							gg.draw(line);
						}
						gg.dispose();
						JFrame frame = new JFrame("Putative matches");
						JLabel label = new JLabel(new ImageIcon(tmp));
						frame.add(label);
						frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
						frame.pack();
						frame.setVisible(true);
					}

					// Use ransac to exclude outliers
					RANSACHomography ransac = RANSACHomography.get(cl);
					int batchCount = 32 * 7 * 16;

					CLBuffer<IntBuffer> randomIndex = cl.createIntBuffer(batchCount * 4);
					CLBuffer<FloatBuffer> A = cl.createFloatBuffer(batchCount * 9 * 9);
					CLBuffer<FloatBuffer> Q = cl.createFloatBuffer(batchCount * 9 * 9);
					CLBuffer<FloatBuffer> t1inv = cl.createFloatBuffer(batchCount * 9);
					CLBuffer<FloatBuffer> t2 = cl.createFloatBuffer(batchCount * 9);
					CLBuffer<FloatBuffer> H = cl.createFloatBuffer(batchCount * 9);

					CLBuffer<FloatBuffer> inliers = cl.createFloatBuffer(batchCount * 2);
					CLBuffer<FloatBuffer> bestH = cl.createFloatBuffer(9);
					CLBuffer<IntBuffer> inlierCount = cl.createIntBuffer(1);
					CLBuffer<FloatBuffer> inlierPoints = cl.createFloatBuffer(pointCount * 4);

					// The random part of ransac - make up sets of 4 random points
					Random rand = new Random(0);
					IntBuffer rb = randomIndex.getBuffer();
					for (int i = 0; i < batchCount; i++) {
						rb.put(rand.nextInt(pointCount));
						rb.put(rand.nextInt(pointCount));
						rb.put(rand.nextInt(pointCount));
						rb.put(rand.nextInt(pointCount));
					}
					rb.rewind();
					q.putWriteBuffer(randomIndex, false);

					System.out.println("ransac");
					ransac.testRANSAC(q, batchCount, surf.getPoints(0), m2permute, randomIndex, A, Q, t1inv, t2, H);
					ransac.testRANSAC(q, batchCount, surf.getPoints(0), m2permute, randomIndex, A, Q, t1inv, t2, H);
					ransac.testRANSAC(q, batchCount, surf.getPoints(0), m2permute, randomIndex, A, Q, t1inv, t2, H);
					System.out.println("test svd");
					ransac.testSVD(q, batchCount, A, Q);
					ransac.testSVD(q, batchCount, A, Q);
					ransac.testSVD(q, batchCount, A, Q);

					System.out.println("inliers");
					ransac.inliers(q, batchCount, pointCount, surf.getPoints(0), m2permute, H, 3, inliers, bestH, inlierCount, inlierPoints);

					q.putReadBuffer(inliers, true);

					FloatBuffer il = inliers.getBuffer();
					float maxc = 0;
					float maxd = 0;
					int maxi = 0;
					for (int i = 0; i < batchCount; i++) {
						float c = il.get(i * 2 + 0);
						float d = il.get(i * 2 + 1);
						//System.out.printf(" %3d: count=%d, distance %8.5f\n", i, (int) c, d);
						if (c > maxc) {
							maxc = c;
							maxd = d;
							maxi = i;
						} else if (c == maxc && d < maxd) {
							maxc = c;
							maxd = d;
							maxi = i;
						}
					}

					q.putReadBuffer(H, true);
					FloatBuffer hb = H.getBuffer();

					{
						int i = maxi;
						System.out.printf("H[%d] =\n", i);
						for (int j = 0; j < 3; j++) {
							for (int k = 0; k < 3; k++) {
								float v = hb.get(i * 9 + j * 3 + k);
								System.out.printf(" %6.3f", v);
							}
							System.out.println();
						}
					}

					{
						q.putReadBuffer(bestH, true);
						hb = bestH.getBuffer();
						System.out.println("bestH=\n");
						for (int j = 0; j < 3; j++) {
							for (int k = 0; k < 3; k++) {
								float v = hb.get(j * 3 + k);
								System.out.printf(" %6.3f", v);
							}
							System.out.println();
						}

					}

					// show inlier matches
					if (true) {
						q.putReadBuffer(inlierCount, false);
						q.putReadBuffer(inlierPoints, true);

						BufferedImage tmp = new BufferedImage(width * 2, height, BufferedImage.TYPE_INT_BGR);
						Graphics2D gg = tmp.createGraphics();

						gg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
						gg.drawImage(img0, 0, 0, null);
						gg.drawImage(img1, width, 0, null);

						Line2D.Float line = new java.awt.geom.Line2D.Float();
						int count = inlierCount.getBuffer().get(0);

						System.out.printf("InlierCount = %d\n", count);
						FloatBuffer ib = inlierPoints.getBuffer();
						for (int i = 0; i < count; i++) {
							line.x1 = ib.get(i * 4 + 0);
							line.y1 = ib.get(i * 4 + 1);
							line.x2 = ib.get(i * 4 + 2) + width;
							line.y2 = ib.get(i * 4 + 3);

							gg.draw(line);
						}
						gg.dispose();
						JFrame frame = new JFrame("RANSAC matches");
						JLabel label = new JLabel(new ImageIcon(tmp));
						frame.add(label);
						frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
						frame.pack();
						frame.setVisible(true);
					}


					q.finish();

					/*
					 long now = System.nanoTime();

					 surf.loadImage(q, id, clsrc0);

					 q.finish();
					 now = System.nanoTime() - now;

					 System.out.printf("Found %d features in %d.%06ds\n", surf.getPointCount(id), now / 1000000000L, (now / 1000) % 1000000);

					 // display features
					 q.putReadBuffer(surf.getLaplacian(id), false);
					 q.putReadBuffer(surf.getOrientations(id), false);
					 q.putReadBuffer(surf.getPoints(id), false);
					 q.putReadBuffer(surf.getScale(id), true);

					 BufferedImage dst = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
					 Graphics2D gg = dst.createGraphics();
					 gg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

					 gg.drawImage(img0, 0, 0, null);
					 drawFeatures(gg, surf, id);

					 // display results
					 if (true) {
					 JFrame frame = new JFrame();
					 JLabel label = new JLabel(new ImageIcon(dst));
					 frame.add(label);
					 frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
					 frame.pack();
					 frame.setVisible(true);
					 }
					 System.out.println("done");
					 */
				} catch (IOException ex) {
					Logger.getLogger(ConvolveXY.class.getName()).log(Level.SEVERE, null, ex);
				}
			}
		};
                SwingUtilities.invokeLater(t);
                t.join();
            }
	}

	public static void drawFeatures(Graphics2D gg, SURFDetector surf, int id) {
		int count = surf.getPointCount(id);
		FloatBuffer points = surf.getPoints(id).getBuffer();
		IntBuffer laplacians = surf.getLaplacian(id).getBuffer();
		FloatBuffer orientations = surf.getOrientations(id).getBuffer();
		FloatBuffer scales = surf.getScale(id).getBuffer();
		Line2D.Float line = new java.awt.geom.Line2D.Float();
		Ellipse2D.Float circle = new java.awt.geom.Ellipse2D.Float();

		for (int i = 0; i < count; i++) {
			float scale = scales.get(i);
			float orientation = orientations.get(i);
			float ix = points.get(i * 2 + 0);
			float iy = points.get(i * 2 + 1);
			int laplacian = laplacians.get(i);

			float s, o;
			float r1, c1, r2, c2, lap;

			s = ((9.0f / 1.2f) * scale) / 3.0f;
			o = orientation;
			lap = laplacian;
			r1 = (iy);
			c1 = (ix);
			c2 = (float) (s * cos(o)) + c1;
			r2 = (float) (s * sin(o)) + r1;

			gg.setColor(Color.green);
			if (o != 0) {// Green line indicates orientation
				line.x1 = c1;
				line.y1 = r1;
				line.x2 = c2;
				line.y2 = r2;
				gg.draw(line);
			} else {// Green dot if using upright version
				circle.setFrameFromCenter(c1, r1, c1 + 1, r1 + 1);
				gg.draw(circle);
			}
			if (lap == 1) { // Blue circles indicate light blobs on dark backgrounds
				gg.setColor(Color.blue);

			} else { // Red circles indicate light blobs on dark backgrounds
				gg.setColor(Color.red);
			}
			circle.setFrameFromCenter(c1, r1, c1 + s, r1 + s);
			gg.draw(circle);
		}
	}
}
